@@ -3,7 +3,7 @@
  *              that either displays or sets the characteristics of
  *              one or more of the system's networking interfaces.
  *
- * Version:     $Id: ifconfig.c,v 1.39 2000/06/11 15:44:21 pb Exp $
+ * Version:     $Id: ifconfig.c,v 1.50 2001/04/13 18:25:18 pb Exp $
  *
  * Author:      Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
  *              and others.  Copyright 1993 MicroWalt Corporation
@@ -20,6 +20,9 @@
  * {1.34} - 19980630 - Arnaldo Carvalho de Melo <acme@conectiva.com.br>
  *                     - gettext instead of catgets for i18n
  *          10/1998  - Andi Kleen. Use interface list primitives.       
+ *	    20001008 - Bernd Eckenfels, Patch from RH for setting mtu 
+ *			(default AF was wrong)
+ *          20010404 - Arnaldo Carvalho de Melo, use setlocale
  */
 
 #define DFLT_AF "inet"
@@ -82,7 +85,7 @@ struct in6_ifreq {
 #include "sockets.h"
 #include "util.h"
 
-char *Release = RELEASE, *Version = "ifconfig 1.40 (2000-05-21)";
+char *Release = RELEASE, *Version = "ifconfig 1.42 (2001-04-13)";
 
 int opt_a = 0;			/* show all interfaces          */
 int opt_i = 0;			/* show the statistics          */
@@ -91,14 +94,11 @@ int opt_v = 0;			/* debugging output flag        */
 int addr_family = 0;		/* currently selected AF        */
 
 /* for ipv4 add/del modes */
-static int get_nmbc_parent(char *parent,
-			unsigned long *nm, unsigned long *bc);
-static int set_ifstate(char *parent,
-			unsigned long ip,
-			unsigned long nm,
-			unsigned long bc,
-			int flag);
-
+static int get_nmbc_parent(char *parent, unsigned long *nm, 
+			   unsigned long *bc);
+static int set_ifstate(char *parent, unsigned long ip,
+		       unsigned long nm, unsigned long bc,
+		       int flag);
 
 static int if_print(char *ifname)
 {
@@ -112,7 +112,7 @@ static int if_print(char *ifname)
     } else {
 	struct interface *ife;
 
-	ife = lookup_interface(ifname,1);
+	ife = lookup_interface(ifname);
 	res = do_if_fetch(ife); 
 	if (res >= 0) 
 	    ife_print(ife);
@@ -125,13 +125,13 @@ static int set_flag(char *ifname, short flag)
 {
     struct ifreq ifr;
 
-    strcpy(ifr.ifr_name, ifname);
+    safe_strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
     if (ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0) {
 	fprintf(stderr, _("%s: unknown interface: %s\n"), 
 		ifname,	strerror(errno));
 	return (-1);
     }
-    strcpy(ifr.ifr_name, ifname);
+    safe_strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
     ifr.ifr_flags |= flag;
     if (ioctl(skfd, SIOCSIFFLAGS, &ifr) < 0) {
 	perror("SIOCSIFFLAGS");
@@ -157,13 +157,13 @@ static int clr_flag(char *ifname, short flag)
     } else
         fd = skfd;
 
-    strcpy(ifr.ifr_name, ifname);
+    safe_strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
     if (ioctl(fd, SIOCGIFFLAGS, &ifr) < 0) {
 	fprintf(stderr, _("%s: unknown interface: %s\n"), 
 		ifname, strerror(errno));
 	return -1;
     }
-    strcpy(ifr.ifr_name, ifname);
+    safe_strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
     ifr.ifr_flags &= ~flag;
     if (ioctl(fd, SIOCSIFFLAGS, &ifr) < 0) {
 	perror("SIOCSIFFLAGS");
@@ -208,7 +208,7 @@ static void usage(void)
 static void version(void)
 {
     fprintf(stderr, "%s\n%s\n", Release, Version);
-    exit(1);
+    exit(0);
 }
 
 static int set_netmask(int skfd, struct ifreq *ifr, struct sockaddr *sa)
@@ -248,16 +248,11 @@ int main(int argc, char **argv)
 #endif
 
 #if I18N
+    setlocale (LC_ALL, "");
     bindtextdomain("net-tools", "/usr/share/locale");
     textdomain("net-tools");
 #endif
 
-    /* Create a channel to the NET kernel. */
-    if ((skfd = sockets_open(0)) < 0) {
-	perror("socket");
-	exit(1);
-    }
-
     /* Find any options. */
     argc--;
     argv++;
@@ -280,8 +275,9 @@ int main(int argc, char **argv)
 	    usage();
 
 	else {
-	    fprintf(stderr, "ifconfig: unknown option %s\n", argv[0]);
-	    fprintf(stderr, "(Use --help for usage information.\n)");
+	    fprintf(stderr, _("ifconfig: option `%s' not recognised.\n"), 
+		    argv[0]);
+	    fprintf(stderr, _("ifconfig: `--help' gives usage information.\n"));
 	    exit(1);
 	}
 
@@ -289,6 +285,12 @@ int main(int argc, char **argv)
 	argc--;
     }
 
+    /* Create a channel to the NET kernel. */
+    if ((skfd = sockets_open(0)) < 0) {
+	perror("socket");
+	exit(1);
+    }
+
     /* Do we have to show the current setup? */
     if (argc == 0) {
 	int err = if_print((char *) NULL);
@@ -303,20 +305,18 @@ int main(int argc, char **argv)
 	(void) close(skfd);
 	exit(err < 0);
     }
+
     /* The next argument is either an address family name, or an option. */
-    if ((ap = get_aftype(*spp)) == NULL)
+    if ((ap = get_aftype(*spp)) != NULL)
+	spp++; /* it was a AF name */
+    else 
 	ap = get_aftype(DFLT_AF);
-    else {
-	/* XXX: should print the current setup if no args left, but only 
-	   for this family */
-	spp++;
+	
+    if (ap) {
 	addr_family = ap->af;
+	skfd = ap->fd;
     }
 
-    if (sockets_open(addr_family) < 0) {
-	perror("family socket");
-	exit(1);
-    }
     /* Process the remaining arguments. */
     while (*spp != (char *) NULL) {
 	if (!strcmp(*spp, "arp")) {
@@ -356,12 +356,13 @@ int main(int argc, char **argv)
 		    goterr = 1;
 		} else {
 		    if (ioctl(skfd, SIOCGIFMAP, &ifr) < 0) {
+			perror("port: SIOCGIFMAP"); 
 			goterr = 1;
 			continue;
 		    }
 		    ifr.ifr_map.port = newport;
 		    if (ioctl(skfd, SIOCSIFMAP, &ifr) < 0) {
-			perror("SIOCSIFMAP");
+			perror("port: SIOCSIFMAP");
 			goterr = 1;
 		    }
 		}
@@ -562,12 +563,14 @@ int main(int argc, char **argv)
 	    if (*++spp == NULL)
 		usage();
 	    if (ioctl(skfd, SIOCGIFMAP, &ifr) < 0) {
+		fprintf(stderr, "mem_start: SIOCGIFMAP: %s\n", strerror(errno)); 
+		spp++; 
 		goterr = 1;
 		continue;
 	    }
 	    ifr.ifr_map.mem_start = strtoul(*spp, NULL, 0);
 	    if (ioctl(skfd, SIOCSIFMAP, &ifr) < 0) {
-		fprintf(stderr, "SIOCSIFMAP: %s\n", strerror(errno));
+		fprintf(stderr, "mem_start: SIOCSIFMAP: %s\n", strerror(errno));
 		goterr = 1;
 	    }
 	    spp++;
@@ -577,12 +580,14 @@ int main(int argc, char **argv)
 	    if (*++spp == NULL)
 		usage();
 	    if (ioctl(skfd, SIOCGIFMAP, &ifr) < 0) {
+		fprintf(stderr, "io_addr: SIOCGIFMAP: %s\n", strerror(errno)); 
+		spp++; 
 		goterr = 1;
 		continue;
 	    }
 	    ifr.ifr_map.base_addr = strtol(*spp, NULL, 0);
 	    if (ioctl(skfd, SIOCSIFMAP, &ifr) < 0) {
-		fprintf(stderr, "SIOCSIFMAP: %s\n", strerror(errno));
+		fprintf(stderr, "io_addr: SIOCSIFMAP: %s\n", strerror(errno));
 		goterr = 1;
 	    }
 	    spp++;
@@ -592,12 +597,14 @@ int main(int argc, char **argv)
 	    if (*++spp == NULL)
 		usage();
 	    if (ioctl(skfd, SIOCGIFMAP, &ifr) < 0) {
+		fprintf(stderr, "irq: SIOCGIFMAP: %s\n", strerror(errno)); 
 		goterr = 1;
+		spp++; 
 		continue;
 	    }
 	    ifr.ifr_map.irq = atoi(*spp);
 	    if (ioctl(skfd, SIOCSIFMAP, &ifr) < 0) {
-		fprintf(stderr, "SIOCSIFMAP: %s\n", strerror(errno));
+		fprintf(stderr, "irq: SIOCSIFMAP: %s\n", strerror(errno));
 		goterr = 1;
 	    }
 	    spp++;
@@ -636,6 +643,12 @@ int main(int argc, char **argv)
 		usage();
 	    if ((hw = get_hwtype(*spp)) == NULL)
 		usage();
+	    if (hw->input == NULL) {
+	    	fprintf(stderr, _("hw address type `%s' has no handler to set address. failed.\n"), *spp);
+	    	spp+=2;
+	    	goterr = 1;
+	    	continue;
+	    }
 	    if (*++spp == NULL)
 		usage();
 	    safe_strncpy(host, *spp, (sizeof host));
@@ -903,9 +916,14 @@ int main(int argc, char **argv)
 		break;
 	    }
 	}
+	if (ap->input == NULL) {
+	   fprintf(stderr, _("ifconfig: Cannot set address for this protocol family.\n"));
+	   exit(1);
+	}
 	if (ap->input(0, host, &sa) < 0) {
 	    ap->herror(host);
-	    usage();
+	    fprintf(stderr, _("ifconfig: `--help' gives usage information.\n"));
+	    exit(1);
 	}
 	memcpy((char *) &ifr.ifr_addr, (char *) &sa, sizeof(struct sockaddr));
 	{
@@ -941,6 +959,7 @@ int main(int argc, char **argv)
 		goterr = 1;
 	    }
 	}
+
        /*
         * Don't do the set_flag() if the address is an alias with a - at the
         * end, since it's deleted already! - Roman
@@ -963,121 +982,128 @@ int main(int argc, char **argv)
 
     return (goterr);
 }
+
 struct ifcmd {
-	int flag;
-	unsigned long addr;
-	char *base;
-	int baselen;
+    int flag;
+    unsigned long addr;
+    char *base;
+    int baselen;
 };
+
 static unsigned char searcher[256];
 
 static int set_ip_using(const char *name, int c, unsigned long ip)
 {
-	struct ifreq ifr;
-	struct sockaddr_in sin;
+    struct ifreq ifr;
+    struct sockaddr_in sin;
 
-	strcpy(ifr.ifr_name, name);
-	memset(&sin, 0, sizeof(struct sockaddr));
-	sin.sin_family = AF_INET;
-	sin.sin_addr.s_addr = ip;
-	memcpy(&ifr.ifr_addr, &sin, sizeof(struct sockaddr));
-	if (ioctl(skfd, c, &ifr) < 0)
-		return -1;
-	return 0;
+    safe_strncpy(ifr.ifr_name, name, IFNAMSIZ);
+    memset(&sin, 0, sizeof(struct sockaddr));
+    sin.sin_family = AF_INET;
+    sin.sin_addr.s_addr = ip;
+    memcpy(&ifr.ifr_addr, &sin, sizeof(struct sockaddr));
+    if (ioctl(skfd, c, &ifr) < 0)
+	return -1;
+    return 0;
 }
+
 static int do_ifcmd(struct interface *x, struct ifcmd *ptr)
 {
-	char *z, *e;
-	struct sockaddr_in *sin;
-	int i;
-
-	if (do_if_fetch(x) < 0)
-		return 0;
-	if (strncmp(x->name, ptr->base, ptr->baselen) != 0)
-		return 0; /* skip */
-	z = strchr(x->name, ':');
-	if (!z || !*z)
-		return 0;
-	z++;
-	for (e = z; *e; e++)
-		if (*e == '-') /* deleted */
-			return 0;
-	i = atoi(z);
-	searcher[i] = 1;
-
-	/* copy */
-	sin = (struct sockaddr_in *)&x->dstaddr;
-	if (sin->sin_addr.s_addr != ptr->addr) {
-		return 0;
-	}
-	
-	if (ptr->flag) {
-		/* turn UP */
-		if (set_flag(x->name, IFF_UP | IFF_RUNNING) == -1)
-			return -1;
-	} else {
-		/* turn DOWN */
-		if (clr_flag(x->name, IFF_UP) == -1)
-			return -1;
-	}
-
-	return 1; /* all done! */
+    char *z, *e;
+    struct sockaddr_in *sin;
+    int i;
+    
+    if (do_if_fetch(x) < 0)
+	return 0;
+    if (strncmp(x->name, ptr->base, ptr->baselen) != 0)
+	return 0; /* skip */
+    z = strchr(x->name, ':');
+    if (!z || !*z)
+	return 0;
+    z++;
+    for (e = z; *e; e++)
+	if (*e == '-') /* deleted */
+	    return 0;
+    i = atoi(z);
+    if (i < 0 || i > 255)
+	abort();
+    searcher[i] = 1;
+    
+    /* copy */
+    sin = (struct sockaddr_in *)&x->dstaddr;
+    if (sin->sin_addr.s_addr != ptr->addr) {
+	return 0;
+    }
+    
+    if (ptr->flag) {
+	/* turn UP */
+	if (set_flag(x->name, IFF_UP | IFF_RUNNING) == -1)
+	    return -1;
+    } else {
+	/* turn DOWN */
+	if (clr_flag(x->name, IFF_UP) == -1)
+	    return -1;
+    }
+    
+    return 1; /* all done! */
 }
 
 
 static int get_nmbc_parent(char *parent,
 			   unsigned long *nm, unsigned long *bc)
 {
-	struct interface *i;
-	struct sockaddr_in *sin;
-
-	i = lookup_interface(parent, 1);
-	if (!i)
-		return -1;
-	if (do_if_fetch(i) < 0)
-		return 0;
-	sin = (struct sockaddr_in *)&i->netmask;
-	memcpy(nm, &sin->sin_addr.s_addr, sizeof(unsigned long));
-	sin = (struct sockaddr_in *)&i->broadaddr;
-	memcpy(bc, &sin->sin_addr.s_addr, sizeof(unsigned long));
+    struct interface *i;
+    struct sockaddr_in *sin;
+    
+    i = lookup_interface(parent);
+    if (!i)
+	return -1;
+    if (do_if_fetch(i) < 0)
 	return 0;
+    sin = (struct sockaddr_in *)&i->netmask;
+    memcpy(nm, &sin->sin_addr.s_addr, sizeof(unsigned long));
+    sin = (struct sockaddr_in *)&i->broadaddr;
+    memcpy(bc, &sin->sin_addr.s_addr, sizeof(unsigned long));
+    return 0;
 }
 
 static int set_ifstate(char *parent, unsigned long ip,
 		       unsigned long nm, unsigned long bc,
 		       int flag)
 {
-	char buf[IFNAMSIZ];
-	struct ifcmd pt;
-	int i;
-
-	pt.base = parent;
-	pt.baselen = strlen(parent);
-	pt.addr = ip;
-	pt.flag = flag;
-	memset(searcher, 0, sizeof(searcher));
-	i = for_all_interfaces((int (*)(struct interface *,void *))do_ifcmd, 
-			       &pt);
-	if (i == -1)
-		return -1;
-	if (i == 1)
-		return 0;
-
-	/* add a new interface */
-	for (i = 0; i < 256; i++)
-		if (searcher[i] == 0)
-			break;
-	if (i == 256)
-		return -1; /* FAILURE!!! out of ip addresses */
-
-	sprintf(buf, "%s:%d", parent, i);
-	if (set_ip_using(buf, SIOCSIFADDR, ip) == -1)
-		return -1;
-	if (set_ip_using(buf, SIOCSIFNETMASK, nm) == -1)
-		return -1;
-	if (set_ip_using(buf, SIOCSIFBRDADDR, bc) == -1)
-		return -1;
-	if (set_flag(buf, IFF_BROADCAST) == -1)
-		return -1;
+    char buf[IFNAMSIZ];
+    struct ifcmd pt;
+    int i;
+    
+    pt.base = parent;
+    pt.baselen = strlen(parent);
+    pt.addr = ip;
+    pt.flag = flag;
+    memset(searcher, 0, sizeof(searcher));
+    i = for_all_interfaces((int (*)(struct interface *,void *))do_ifcmd, 
+			   &pt);
+    if (i == -1)
+	return -1;
+    if (i == 1)
 	return 0;
+    
+    /* add a new interface */
+    for (i = 0; i < 256; i++)
+	if (searcher[i] == 0)
+	    break;
+
+    if (i == 256)
+	return -1; /* FAILURE!!! out of ip addresses */
+    
+    if (snprintf(buf, IFNAMSIZ, "%s:%d", parent, i) > IFNAMSIZ)
+	return -1;
+    if (set_ip_using(buf, SIOCSIFADDR, ip) == -1)
+	return -1;
+    if (set_ip_using(buf, SIOCSIFNETMASK, nm) == -1)
+	return -1;
+    if (set_ip_using(buf, SIOCSIFBRDADDR, bc) == -1)
+	return -1;
+    if (set_flag(buf, IFF_BROADCAST) == -1)
+	return -1;
+    return 0;
 }
