@@ -1,8 +1,8 @@
 struct user_net_device_stats {
-    unsigned long rx_packets;	/* total packets received       */
-    unsigned long tx_packets;	/* total packets transmitted    */
-    unsigned long rx_bytes;	/* total bytes received         */
-    unsigned long tx_bytes;	/* total bytes transmitted      */
+    unsigned long long rx_packets;	/* total packets received       */
+    unsigned long long tx_packets;	/* total packets transmitted    */
+    unsigned long long rx_bytes;	/* total bytes received         */
+    unsigned long long tx_bytes;	/* total bytes transmitted      */
     unsigned long rx_errors;	/* bad packets received         */
     unsigned long tx_errors;	/* packet transmit problems     */
     unsigned long rx_dropped;	/* no space in linux buffers    */
@@ -28,8 +28,7 @@ struct user_net_device_stats {
 };
 
 struct interface {
-    struct interface *next;
-
+    struct interface *next, *prev; 
     char name[IFNAMSIZ];	/* interface name        */
     short type;			/* if type               */
     short flags;		/* various flags         */
@@ -66,7 +65,7 @@ extern int if_fetch(struct interface *ife);
 
 extern int for_all_interfaces(int (*)(struct interface *, void *), void *);
 extern int free_interface_list(void);
-extern struct interface *lookup_interface(char *name, int readlist);
+extern struct interface *lookup_interface(char *name);
 extern int if_readlist(void);
 
 extern int do_if_fetch(struct interface *ife);
