@@ -4,7 +4,10 @@
    10/1998 partly rewriten by Andi Kleen to support an interface list.   
    I don't claim that the list operations are efficient @).  
 
-   $Id: interface.c,v 1.9 2000/05/21 19:35:34 pb Exp $
+   8/2000  Andi Kleen make the list operations a bit more efficient.
+   People are crazy enough to use thousands of aliases now.
+
+   $Id: interface.c,v 1.14 2001/02/10 19:31:15 pb Exp $
  */
 
 #include "config.h"
@@ -84,38 +87,41 @@ int procnetdev_vsn = 1;
 
 int ife_short;
 
-static struct interface *int_list;
+static struct interface *int_list, *int_last;
 
-void add_interface(struct interface *n)
+static int if_readlist_proc(char *);
+
+static struct interface *add_interface(char *name)
 {
-    struct interface *ife, **pp;
+    struct interface *ife, **nextp, *new;
 
-    pp = &int_list;
-    for (ife = int_list; ife; pp = &ife->next, ife = ife->next) {
-	if (nstrcmp(ife->name, n->name) > 0)
-	    break;
+    for (ife = int_last; ife; ife = ife->prev) {
+	    int n = nstrcmp(ife->name, name); 
+	    if (n == 0) 
+		    return ife; 
+	    if (n < 0) 
+		    break; 
     }
-    n->next = (*pp);
-    (*pp) = n;
+    new(new); 
+    safe_strncpy(new->name, name, IFNAMSIZ); 
+    nextp = ife ? &ife->next : &int_list;
+    new->prev = ife;
+    new->next = *nextp; 
+    if (new->next) 
+	    new->next->prev = new; 
+    else
+	    int_last = new; 
+    *nextp = new; 
+    return new; 
 }
 
-struct interface *lookup_interface(char *name, int readlist)
+struct interface *lookup_interface(char *name)
 {
     struct interface *ife = NULL;
 
-    if (int_list || (readlist && if_readlist() >= 0)) {
-		for (ife = int_list; ife; ife = ife->next) {
-	    	if (!strcmp(ife->name, name))
-			break;
-		}
-    }
-
-    if (!ife) { 
-	new(ife);
-	safe_strncpy(ife->name, name, IFNAMSIZ);  
-	add_interface(ife);
-    }
-
+    if (if_readlist_proc(name) < 0) 
+	    return NULL; 
+    ife = add_interface(name); 
     return ife;
 }
 
@@ -182,7 +188,7 @@ static int if_readconf(void)
 
     ifr = ifc.ifc_req;
     for (n = 0; n < ifc.ifc_len; n += sizeof(struct ifreq)) {
-	lookup_interface(ifr->ifr_name,0);
+	add_interface(ifr->ifr_name);
 	ifr++;
     }
     err = 0;
@@ -233,7 +239,7 @@ static int get_dev_fields(char *bp, struct interface *ife)
     switch (procnetdev_vsn) {
     case 3:
 	sscanf(bp,
-	"%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
+	"%llu %llu %lu %lu %lu %lu %lu %lu %llu %llu %lu %lu %lu %lu %lu %lu",
 	       &ife->stats.rx_bytes,
 	       &ife->stats.rx_packets,
 	       &ife->stats.rx_errors,
@@ -253,7 +259,7 @@ static int get_dev_fields(char *bp, struct interface *ife)
 	       &ife->stats.tx_compressed);
 	break;
     case 2:
-	sscanf(bp, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
+	sscanf(bp, "%llu %llu %lu %lu %lu %lu %llu %llu %lu %lu %lu %lu %lu",
 	       &ife->stats.rx_bytes,
 	       &ife->stats.rx_packets,
 	       &ife->stats.rx_errors,
@@ -271,7 +277,7 @@ static int get_dev_fields(char *bp, struct interface *ife)
 	ife->stats.rx_multicast = 0;
 	break;
     case 1:
-	sscanf(bp, "%ld %ld %ld %ld %ld %ld %ld %ld %ld %ld %ld",
+	sscanf(bp, "%llu %lu %lu %lu %lu %llu %lu %lu %lu %lu %lu",
 	       &ife->stats.rx_packets,
 	       &ife->stats.rx_errors,
 	       &ife->stats.rx_dropped,
@@ -292,13 +298,19 @@ static int get_dev_fields(char *bp, struct interface *ife)
     return 0;
 }
 
-int if_readlist(void)
+static int if_readlist_proc(char *target)
 {
+    static int proc_read; 
     FILE *fh;
     char buf[512];
     struct interface *ife;
     int err;
 
+    if (proc_read) 
+	    return 0; 
+    if (!target) 
+	    proc_read = 1;
+
     fh = fopen(_PATH_PROCNET_DEV, "r");
     if (!fh) {
 		fprintf(stderr, _("Warning: cannot open %s (%s). Limited output.\n"),
@@ -336,22 +348,19 @@ int if_readlist(void)
 
     err = 0;
     while (fgets(buf, sizeof buf, fh)) {
-	char *s;
-
-	new(ife);
-
-	s = get_name(ife->name, buf);
+	char *s, name[IFNAMSIZ];
+	s = get_name(name, buf);    
+	ife = add_interface(name);
 	get_dev_fields(s, ife);
 	ife->statistics_valid = 1;
-
-	add_interface(ife);
+	if (target && !strcmp(target,name))
+		break;
     }
     if (ferror(fh)) {
 	perror(_PATH_PROCNET_DEV);
 	err = -1;
+	proc_read = 0; 
     }
-    if (!err)
-	err = if_readconf();
 
 #if 0
     free(fmt);
@@ -360,6 +369,14 @@ int if_readlist(void)
     return err;
 }
 
+int if_readlist(void) 
+{ 
+    int err = if_readlist_proc(NULL); 
+    if (!err)
+	    err = if_readconf();
+    return err;
+} 
+
 /* Support for fetching an IPX address */
 
 #if HAVE_AFIPX
@@ -563,18 +580,19 @@ int do_if_print(struct interface *ife, void *cookie)
 void ife_print_short(struct interface *ptr)
 {
     printf("%-5.5s ", ptr->name);
-    printf("%5d %3d ", ptr->mtu, ptr->metric);
+    printf("%5d %3d", ptr->mtu, ptr->metric);
     /* If needed, display the interface statistics. */
     if (ptr->statistics_valid) {
-	printf("%8lu %6lu %6lu %6lu ",
+	printf("%8llu %6lu %6lu %6lu",
 	       ptr->stats.rx_packets, ptr->stats.rx_errors,
 	       ptr->stats.rx_dropped, ptr->stats.rx_fifo_errors);
-	printf("%8lu %6lu %6lu %6lu ",
+	printf("%8llu %6lu %6lu %6lu ",
 	       ptr->stats.tx_packets, ptr->stats.tx_errors,
 	       ptr->stats.tx_dropped, ptr->stats.tx_fifo_errors);
     } else {
 	printf("%-56s", _("     - no statistics available -"));
     }
+    /* DONT FORGET TO ADD THE FLAGS IN ife_print_long, too */
     if (ptr->flags == 0)
 	printf(_("[NO FLAGS]"));
     if (ptr->flags & IFF_ALLMULTI)
@@ -585,18 +603,29 @@ void ife_print_short(struct interface *ptr)
 	printf("D");
     if (ptr->flags & IFF_LOOPBACK)
 	printf("L");
-    if (ptr->flags & IFF_PROMISC)
+    if (ptr->flags & IFF_MULTICAST)
 	printf("M");
+#ifdef HAVE_DYNAMIC
+    if (ptr->flags & IFF_DYNAMIC)
+	printf("d");
+#endif
+    if (ptr->flags & IFF_PROMISC)
+	printf("P");
     if (ptr->flags & IFF_NOTRAILERS)
 	printf("N");
     if (ptr->flags & IFF_NOARP)
 	printf("O");
     if (ptr->flags & IFF_POINTOPOINT)
 	printf("P");
+    if (ptr->flags & IFF_SLAVE)
+	printf("s");
+    if (ptr->flags & IFF_MASTER)
+	printf("m");
     if (ptr->flags & IFF_RUNNING)
 	printf("R");
     if (ptr->flags & IFF_UP)
 	printf("U");
+    /* DONT FORGET TO ADD THE FLAGS IN ife_print_long, too */
     printf("\n");
 }
 
@@ -606,6 +635,10 @@ void ife_print_long(struct interface *ptr)
     struct hwtype *hw;
     int hf;
     int can_compress = 0;
+    unsigned long long rx, tx, short_rx, short_tx;
+    char Rext[5]="b";
+    char Text[5]="b";
+
 #if HAVE_AFIPX
     static struct aftype *ipxtype = NULL;
 #endif
@@ -747,6 +780,7 @@ void ife_print_long(struct interface *ptr)
 #endif
 
     printf("          ");
+    /* DONT FORGET TO ADD THE FLAGS IN ife_print_short, too */
     if (ptr->flags == 0)
 	printf(_("[NO FLAGS] "));
     if (ptr->flags & IFF_UP)
@@ -779,7 +813,7 @@ void ife_print_long(struct interface *ptr)
     if (ptr->flags & IFF_DYNAMIC)
 	printf(_("DYNAMIC "));
 #endif
-
+    /* DONT FORGET TO ADD THE FLAGS IN ife_print_short */
     printf(_(" MTU:%d  Metric:%d"),
 	   ptr->mtu, ptr->metric ? ptr->metric : 1);
 #ifdef SIOCSKEEPALIVE
@@ -798,16 +832,24 @@ void ife_print_long(struct interface *ptr)
 	 */
 	printf("          ");
 
-	printf(_("RX packets:%lu errors:%lu dropped:%lu overruns:%lu frame:%lu\n"),
+	printf(_("RX packets:%llu errors:%lu dropped:%lu overruns:%lu frame:%lu\n"),
 	       ptr->stats.rx_packets, ptr->stats.rx_errors,
 	       ptr->stats.rx_dropped, ptr->stats.rx_fifo_errors,
 	       ptr->stats.rx_frame_errors);
 	if (can_compress)
 	    printf(_("             compressed:%lu\n"), ptr->stats.rx_compressed);
 
-	printf("          ");
+	rx = ptr->stats.rx_bytes;  
+	tx = ptr->stats.tx_bytes;
+	short_rx = rx * 10;  
+	short_tx = tx * 10;
+	if (rx > 1048576) { short_rx /= 1048576;  strcpy(Rext, "Mb"); }
+	else if (rx > 1024) { short_rx /= 1024;  strcpy(Rext, "Kb"); }
+	if (tx > 1048576) { short_tx /= 1048576;  strcpy(Text, "Mb"); }
+	else if (tx > 1024) { short_tx /= 1024;  strcpy(Text, "Kb"); }
 
-	printf(_("TX packets:%lu errors:%lu dropped:%lu overruns:%lu carrier:%lu\n"),
+	printf("          ");
+	printf(_("TX packets:%llu errors:%lu dropped:%lu overruns:%lu carrier:%lu\n"),
 	       ptr->stats.tx_packets, ptr->stats.tx_errors,
 	       ptr->stats.tx_dropped, ptr->stats.tx_fifo_errors,
 	       ptr->stats.tx_carrier_errors);
@@ -816,7 +858,12 @@ void ife_print_long(struct interface *ptr)
 	    printf(_("compressed:%lu "), ptr->stats.tx_compressed);
 	if (ptr->tx_queue_len != -1)
 	    printf(_("txqueuelen:%d "), ptr->tx_queue_len);
-	printf("\n");
+	printf("\n          ");
+	printf(_("RX bytes:%llu (%lu.%lu %s)  TX bytes:%llu (%lu.%lu %s)\n"),
+	       rx, (unsigned long)(short_rx / 10), 
+	       (unsigned long)(short_rx % 10), Rext, 
+	       tx, (unsigned long)(short_tx / 10), 
+	       (unsigned long)(short_tx % 10), Text);
     }
 
     if ((ptr->map.irq || ptr->map.mem_start || ptr->map.dma ||
